@@ -35,17 +35,19 @@ class DSDLDefinition:
                  root_namespace_path: str):
         # Normalizing the path and reading the definition text
         self._file_path = os.path.abspath(file_path)
-        root_namespace_path = os.path.abspath(root_namespace_path)
+        del file_path
+        self._root_namespace_path = os.path.abspath(root_namespace_path)
+        del root_namespace_path
         with open(self._file_path) as f:
             self._text = str(f.read())
 
         # Checking the sanity of the root directory path - can't contain separators
-        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in os.path.split(root_namespace_path)[-1]:
-            raise FileNameFormatError('Invalid namespace name', path=root_namespace_path)
+        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in os.path.split(self._root_namespace_path)[-1]:
+            raise FileNameFormatError('Invalid namespace name', path=self._root_namespace_path)
 
         # Determining the relative path within the root namespace directory
-        relative_path = str(os.path.join(os.path.split(root_namespace_path)[-1],
-                                         os.path.relpath(self._file_path, root_namespace_path)))
+        relative_path = str(os.path.join(os.path.split(self._root_namespace_path)[-1],
+                                         os.path.relpath(self._file_path, self._root_namespace_path)))
 
         relative_directory, basename = [str(x) for x in os.path.split(relative_path)]   # type: str, str
 
@@ -125,13 +127,12 @@ class DSDLDefinition:
                      log_prefix, len(lookup_definitions),
                      ', '.join(set(sorted(map(lambda x: x.root_namespace, lookup_definitions)))))
         try:
-            # We have to import this class at function level to break recursive dependency.
-            # Maybe I have messed up the architecture? Should think about it later.
-            from ._data_type_builder import DataTypeBuilder
-            builder = DataTypeBuilder(definition=self,
-                                      lookup_definitions=lookup_definitions,
-                                      print_output_handler=print_output_handler,
-                                      allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id)
+            builder = _data_type_builder.DataTypeBuilder(
+                definition=self,
+                lookup_definitions=lookup_definitions,
+                print_output_handler=print_output_handler,
+                allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id
+            )
             with open(self.file_path) as f:
                 _parser.parse(f.read(), builder)
 
@@ -196,6 +197,10 @@ class DSDLDefinition:
     def file_path(self) -> str:
         return self._file_path
 
+    @property
+    def root_namespace_path(self) -> str:
+        return self._root_namespace_path
+
     def __eq__(self, other: object) -> bool:
         """
         Two definitions will compare equal if they share the same name AND version number.
@@ -211,3 +216,8 @@ class DSDLDefinition:
             (self.full_name, self.version, self.fixed_port_id, self.file_path)
 
     __repr__ = __str__
+
+
+# Moved this import here to break recursive dependency.
+# Maybe I have messed up the architecture? Should think about it later.
+from . import _data_type_builder  # noqa: E402
